"""
Acts as the main entry point for the grascii search command.

This can be invoked as a standalone program:
$ python -m grascii.search --help
"""

from __future__ import annotations

import argparse
import sys
from typing import Iterable, Optional

from grascii import regen
from grascii.dictionary import DictionaryNotFound
from grascii.parser import InvalidGrascii
from grascii.searchers import (
    GrasciiSearcher,
    RegexSearcher,
    ReverseSearcher,
    Searcher,
    SearchResult,
)

SUPPORTS_INTERACTIVE = False
try:
    from grascii.interactive import InteractiveSearcher

    SUPPORTS_INTERACTIVE = True
except ImportError:
    pass


description = "Search a Grascii Dictionary"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    """Configure an ArgumentParser parser to parse the search command-line
    options.

    :param argparser: A fresh ArgumentParser to configure.
    """

    group = argparser.add_mutually_exclusive_group(required=True)
    group.add_argument("-g", "--grascii", help="the grascii string to search for")
    group.add_argument(
        "-e", "--regexp", help="a custom regular expression to use in the search"
    )
    group.add_argument("-r", "--reverse", help="search by word instead")
    group.add_argument(
        "-i", "--interactive", action="store_true", help="run in interactive mode"
    )
    argparser.add_argument(
        "-u",
        "--uncertainty",
        type=int,
        choices=range(3),
        help="the uncertainty of the search term",
    )
    argparser.add_argument(
        "-s",
        "--search-mode",
        choices=[mode.value for mode in regen.SearchMode],
        help="the kind of search to perform",
    )
    argparser.add_argument(
        "-a",
        "--annotation-mode",
        choices=[mode.value for mode in regen.Strictness],
        help="how to handle annotations in grascii",
    )
    argparser.add_argument(
        "-p",
        "--aspirate-mode",
        choices=[mode.value for mode in regen.Strictness],
        help="how to handle aspirates in grascii",
    )
    argparser.add_argument(
        "-j",
        "--disjoiner-mode",
        choices=[mode.value for mode in regen.Strictness],
        help="how to handle disjoiners in grascii",
    )
    argparser.add_argument(
        "-n",
        "--interpretation",
        choices=["best", "all"],
        help="how to handle ambiguous grascii strings",
    )
    argparser.add_argument(
        "-f",
        "--fix-first",
        action="store_true",
        help="apply an uncertainty of 0 to the first token",
    )
    argparser.add_argument(
        "-d",
        "--dictionary",
        dest="dictionaries",
        action="append",
        help="add a dictionary to be searched",
    )
    argparser.add_argument(
        "--no-sort",
        action="store_true",
        help="do not sort the search results",
    )


def search(**kwargs) -> Optional[Iterable[SearchResult]]:
    """Run a grascii dictionary search. Parameters can consist of
    any parameters used by the search method of any subclass of
    Searcher. One, and only one, of the parameters list below
    is required.

    :param grascii: A grascii string to use in a search.
    :param interactive: A flag enabling an interactive search.
    :param reverse: A word to search for in the dictionary.
    :param regexp: A regular expression to use in a search.
    :type grascii: str
    :type interactive: bool
    :type reverse: str
    :type regexp: str

    :returns: A list of search results, or None if run in interactive mode
    """

    searcher: Searcher
    if kwargs.get("grascii"):
        searcher = GrasciiSearcher(**kwargs)
    elif kwargs.get("interactive"):
        if not SUPPORTS_INTERACTIVE:
            print("The interactive extra is not installed", file=sys.stderr)
            return None
        searcher = InteractiveSearcher(**kwargs)
    elif kwargs.get("reverse"):
        searcher = ReverseSearcher(**kwargs)
    else:
        searcher = RegexSearcher(**kwargs)
    if kwargs.get("no_sort"):
        return searcher.search(**kwargs)
    return searcher.sorted_search(**kwargs)


def cli_search(args: argparse.Namespace) -> None:
    """Run a search using arguments parsed from the command line.

    :param args: A namespace of parsed arguments.
    """

    try:
        results = search(**{k: v for k, v in vars(args).items() if v is not None})
    except InvalidGrascii as e:
        print("Invalid Grascii String", file=sys.stderr)
        print(e.context, file=sys.stderr)
        return
    except DictionaryNotFound as e:
        print("Dictionary Not Found", file=sys.stderr)
        print(e.name, file=sys.stderr)
        return
    else:
        if results is not None:
            count = 0
            for result in results:
                print(result.entry.grascii, result.entry.translation)
                count += 1
            print("Results:", count)


def main() -> None:
    """Run a search using arguments retrieved from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_search(args)
    search(**{k: v for k, v in vars(args).items() if v is not None})


if __name__ == "__main__":
    main()



"""
Contains an implementation of an Interactive Searcher for an
interactive cli experience.
"""

from __future__ import annotations

try:
    import questionary
    from questionary.prompts.common import Choice
except ImportError:
    raise ImportError("Grascii: interactive extra dependencies are not installed")

import sys
from typing import Callable, Iterable, List, Optional, Sequence, Tuple, TypeVar

from grascii import metrics, regen
from grascii.dictionary import Dictionary
from grascii.dictionary.list import get_built_ins, get_installed
from grascii.parser import Interpretation, InvalidGrascii, interpretation_to_string
from grascii.searchers import GrasciiSearcher

T = TypeVar("T")


class InteractiveSearcher(GrasciiSearcher):
    """This subclass of GrasciiSearcher runs an interactive search
    experience for performing grascii searches with support for changing
    search parameters.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.available_dicts = set(self.dictionaries)
        installed = map(lambda s: Dictionary.new(s), get_installed())
        built_ins = map(lambda s: Dictionary.new(s), get_built_ins())
        self.available_dicts.update(installed, built_ins)

    def search(self, **kwargs):
        """
        :param grascii: [Required] The grascii string to use in the search.
        :param uncertainty: The uncertainty of the grascii string.
        :param search_mode: The search mode to use.
        :param annotation_mode: How to handle annotations in the search.
        :param aspirate_mode: How to handle annotations in the search.
        :param disjoiner_mode: How to handle annotations in the search.
        :param fix_first: Apply an uncertainty of 0 to the first token.
        :type grascii: str
        :type uncertainty: int: 0, 1, or 2
        :type search_mode: str: one of regen.SearchMode values
        :type annotation_mode: one of regen.Strictness values
        :type aspirate_mode: one of regen.Strictness values
        :type disjoiner_mode: one of regen.Strictness values
        :type fix_first: bool
        :returns: None
        :rtype: None
        """

        self.extract_search_args(**kwargs)
        self.sort = not kwargs.get("no_sort")
        self._run_interactive()

    def _choose_interpretation(
        self, interpretations: Sequence[Interpretation]
    ) -> Optional[int]:
        """Prompt the user to choose an interpretation(s).

        :param interpretations: A collection of Interpretations to present
            to the user
        :returns: The index of the selected item in interpretations.
        """

        if len(interpretations) == 1:
            return 0
        prompt = "Choose an interpretation to use in the search:"
        choices = [Choice(title="all", value=0)]
        for i, interp in enumerate(interpretations):
            choices.append(Choice(title=interpretation_to_string(interp), value=i + 1))
        return questionary.select(prompt, choices).ask()

    def _run_interactive(self) -> None:
        """Run an interactive search loop."""

        previous_search = None
        while True:
            action = questionary.select(
                "What would you like to do?",
                [
                    "New Search",
                    Choice("Modify Search", disabled=not previous_search),
                    "Edit Settings",
                    "Exit",
                ],
            ).ask()

            if action is None or action == "Exit":
                exit()
            elif action == "New Search":
                previous_search = self._interactive_search()
            elif action == "Modify Search":
                previous_search = self._interactive_search(previous_search)
            elif action == "Edit Settings":
                while True:
                    action = questionary.select(
                        "Search Settings",
                        [
                            Choice(
                                title="Uncertainty [{}]".format(self.uncertainty),
                                value=1,
                            ),
                            Choice(
                                title="Search Mode [{}]".format(self.search_mode.value),
                                value=2,
                            ),
                            Choice(
                                title="Annotation Mode [{}]".format(
                                    self.annotation_mode.value
                                ),
                                value=3,
                            ),
                            Choice(
                                title="Aspirate Mode [{}]".format(
                                    self.aspirate_mode.value
                                ),
                                value=4,
                            ),
                            Choice(
                                title="Disjoiner Mode [{}]".format(
                                    self.disjoiner_mode.value
                                ),
                                value=5,
                            ),
                            Choice(
                                title="Fix First [{}]".format(self.fix_first), value=6
                            ),
                            Choice(
                                title="Dictionaries [{} selected]".format(
                                    len(self.dictionaries)
                                ),
                                value=7,
                            ),
                            "Back",
                        ],
                    ).ask()

                    if action is None or action == "Back":
                        break
                    elif action == 1:
                        self._change_arg(
                            "uncertainty", range(3), display_name="Uncertainty"
                        )
                    elif action == 2:
                        self._change_arg(
                            "search_mode",
                            list(regen.SearchMode),
                            convert=self._get_enum_value,
                            display_name="Search mode",
                        )
                    elif action == 3:
                        self._change_arg(
                            "annotation_mode",
                            list(regen.Strictness),
                            convert=self._get_enum_value,
                            display_name="Annotation mode",
                        )
                    elif action == 4:
                        self._change_arg(
                            "aspirate_mode",
                            list(regen.Strictness),
                            convert=self._get_enum_value,
                            display_name="Aspirate mode",
                        )
                    elif action == 5:
                        self._change_arg(
                            "disjoiner_mode",
                            list(regen.Strictness),
                            convert=self._get_enum_value,
                            display_name="Disjoiner mode",
                        )
                    elif action == 6:
                        self._change_arg(
                            "fix_first", [True, False], display_name="Fix First"
                        )
                    elif action == 7:
                        self._select_dictionaries()

    def _get_enum_value(self, enum):
        return enum.value

    def _change_arg(
        self,
        arg_name,
        options: Iterable[T],
        display_name: str = None,
        convert: Callable[[T], str] = str,
    ) -> None:
        """Prompt the user to select the value of a search parameter and set
            the new value.

        :param arg_name: The name of the argument to change.
        :param options: A collection of choices to present to the user.
        :param display_name: The title of the prompt.
        :param convert: A function that returns a string given an object of
            type T.
        """

        choices = []
        for option in options:
            selected = getattr(self, arg_name) == option
            choices.append(
                Choice(title=convert(option), value=option, checked=selected)
            )
        title = display_name if display_name else "Set " + arg_name
        setting = questionary.select(title, choices).ask()
        if setting is not None:
            setattr(self, arg_name, setting)

    def _select_dictionaries(self) -> None:
        choices = []
        for dictionary in self.available_dicts:
            selected = dictionary in self.dictionaries
            choices.append(
                Choice(title=dictionary.name, value=dictionary, checked=selected)
            )
        title = "Choose Dictionaries"
        dictionaries = questionary.checkbox(
            title,
            choices,
            validate=lambda d: True if len(d) > 0 else "Select at least one dictionary",
        ).ask()
        if dictionaries:
            self.dictionaries = dictionaries

    def _interactive_search(self, previous: str = None) -> Optional[str]:
        """Run an interactive search.

        :param previous: The previous search performed in this interactive
            session.
        :returns: The search string used.
        """

        search, interpretations = self._get_grascii_search(previous)
        if search is None:
            return previous
        index = self._choose_interpretation(interpretations)
        if index is None:
            return search
        builder = regen.RegexBuilder(
            uncertainty=self.uncertainty,
            search_mode=self.search_mode,
            fix_first=self.fix_first,
            annotation_mode=self.annotation_mode,
            aspirate_mode=self.aspirate_mode,
            disjoiner_mode=self.disjoiner_mode,
        )
        if index == 0:
            interps = interpretations
        else:
            interps = interpretations[index - 1 : index]
        patterns = builder.generate_patterns_map(interps)
        starting_letters = builder.get_starting_letters(interps)
        results = self.perform_search(patterns, starting_letters)
        if self.sort:
            results = sorted(results, key=lambda r: metrics.grascii_standard(r))
        count = 0
        display_all = False
        for result in results:
            action = "Next"
            if not display_all:
                action = questionary.select(
                    "Search Results", ["Next", "Display All", "End Search"]
                ).ask()
            if action is None or action == "End Search":
                break
            elif action == "Display All":
                display_all = True
            print(result.entry.grascii, result.entry.translation)
            count += 1

        print("Results:", count)
        print()
        return search

    def _get_grascii_search(
        self, previous: Optional[str] = None
    ) -> Tuple[Optional[str], Optional[List[Interpretation]]]:
        """Prompt the user for a grascii string.

        :param previous: The previous grascii string used in a search.
        :returns: A grascii string and associated interpretations
        """

        while True:
            if previous:
                search = questionary.text("Enter Search:", default=previous).ask()
            else:
                search = questionary.text("Enter Search:").ask()

            if search is None:
                return search, None
            if search == "":
                continue
            search = search.upper()
            try:
                interpretations = list(self._parser.interpret(search))
            except InvalidGrascii as e:
                print("Invalid Grascii String", file=sys.stderr)
                print(e.context, file=sys.stderr)
                previous = search
                continue
            return search, interpretations



from __future__ import annotations

from platformdirs import user_data_path

from grascii import APP_NAME

INSTALLATION_DIR = user_data_path(APP_NAME) / "dictionaries"

BUILTINS_PACKAGE = "grascii.dictionary"


class DictionaryException(Exception):
    """The base class for all dictionary-related exceptions."""

    pass


class DictionaryAlreadyExists(DictionaryException):
    def __init__(self, name: str) -> None:
        self.name = name


class DictionaryNotFound(DictionaryException):
    def __init__(self, name: str) -> None:
        self.name = name


def get_dictionary_installed_name(name: str) -> str:
    """Get the installed name of a dictionary.

    i.e. Prefixed with ':'
    """
    if not name:
        raise ValueError("name cannot be the empty string")
    if name[0] == ":":
        if len(name) == 1:
            raise ValueError("name cannot be ':'")
        return name
    return ":" + name


def get_dictionary_path_name(name: str) -> str:
    """Get the path name of a dictionary.

    i.e. Not prefixed with ':'
    """
    if not name:
        raise ValueError("name cannot be the empty string")
    if name[0] == ":":
        if len(name) == 1:
            raise ValueError("name cannot be ':'")
        return name[1:]
    return name


def is_dictionary_installed_name(name: str) -> bool:
    """Check whether the given name represents an installed dictionary."""
    return len(name) > 1 and name[0] == ":"



#!/usr/bin/python

from __future__ import annotations

import sys
from pathlib import Path

from setuptools import setup
from setuptools.command.build_py import build_py

sys.path.insert(0, str(Path(__file__).parent))

from scripts.build_dictionaries import build_dictionaries  # noqa: E402
from scripts.create_grascii_regex import create_grascii_regex  # noqa: E402


class PreBuild(build_py):
    def run(self):
        build_dictionaries()
        create_grascii_regex()
        build_py.run(self)


setup(cmdclass={"build_py": PreBuild})



from __future__ import annotations

import argparse
import sys
from pathlib import Path
from shutil import copy
from typing import Optional

from grascii.dictionary.common import (
    INSTALLATION_DIR,
    DictionaryAlreadyExists,
    get_dictionary_installed_name,
    get_dictionary_path_name,
)

description = "Install a Grascii Dictionary"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    argparser.add_argument(
        "-n",
        "--name",
        action="store",
        help="The name to give to the installed dictionary.",
    )
    argparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Allow overwriting of an existing installed dictionary.",
    )
    argparser.add_argument(
        "dictionary",
        action="store",
        type=Path,
        help="Path to the built dictionary to install.",
    )


def install_dictionary(
    dictionary: Path,
    install_dir: Path,
    name: Optional[str] = None,
    force: bool = False,
) -> str:
    """Install a dictionary to an installation directory.

    :param dictionary: A path to the output directory of a dictionary build.
    :param install_dir: A path to install the dictionary to.
    :param name: An optional name to give the installed dictionary.
    :param force: If True, an existing dictionary can be overwritten.
    :type dictionary: Path
    :type install_dir: Path
    :type name: str
    :type force: bool

    :returns: The installed name of the dictionary.
    """
    if name is None:
        name = dictionary.name
    else:
        name = get_dictionary_path_name(name)
    destination = install_dir / name
    if destination.exists() and not force:
        raise DictionaryAlreadyExists(name)
    destination.mkdir(parents=True, exist_ok=True)
    files = dictionary.glob("[A-Z]")
    for f in files:
        copy(f, destination)
    return get_dictionary_installed_name(name)


def cli_install(args: argparse.Namespace) -> None:
    try:
        name = install_dictionary(
            args.dictionary, INSTALLATION_DIR, args.name, args.force
        )
    except DictionaryAlreadyExists as e:
        print("A dictionary named", e.name, "already exists.", file=sys.stderr)
        print("Provide a different name with --name.", file=sys.stderr)
        print("If you would like to overwrite it, run with --force.", file=sys.stderr)
    else:
        print("Successfully installed", args.dictionary, "as", name)


def main() -> None:
    """Run the install command using arguments from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_install(args)


if __name__ == "__main__":
    main()



"""
MIT License

Copyright (c) 2022 chanicpanic

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import sys
from functools import reduce
from itertools import chain, product, repeat
from typing import Any, Collection, Iterable, Iterator, Tuple, TypeVar

from lark import Tree, Transformer
from lark.visitors import Interpreter

if sys.version_info >= (3, 8):
    from math import prod
else:

    def prod(nums, start=1):
        return reduce(lambda x, y: x * y, nums, start)


T = TypeVar("T")


class CountedTree(Tree):
    """A tree subclass with an additional attribute, ``derivation_count``, that
    represents the total number of possible derivations of this tree.
    Derivations are counted based on the number and position of '_ambig' nodes.

    Take caution using the constructor directly. All tree children must be
    instances of ``CountedTree`` for the derivation count to be accurate.
    Changing the children after construction will not update the derivation count.
    """
    def __init__(self, data, children, meta=None):
        super().__init__(data, children, meta)
        derivation_counts = map(_get_derivation_count, children)
        self.derivation_count = (sum(derivation_counts) if data == "_ambig" else prod(derivation_counts))


class CountTrees(Transformer):
    """A transformer that transforms a ``Tree`` into a ``CountedTree``."""
    def __default__(self, data, children, meta):
        return CountedTree(data, children, meta)


def _get_derivation_count(tree: Any) -> int:
    return getattr(tree, "derivation_count", 1)


def _repeat_each(iterable: Iterable[T], n: int) -> Iterator[T]:
    """Repeat each element of the iterable n times.

    Recipe from more-itertools: https://github.com/more-itertools/more-itertools
    """
    return chain.from_iterable(map(repeat, iterable, repeat(n)))


def _ncycles(iterable: Iterable[T], n: int) -> Iterator[T]:
    """Return the elements of the iterable n times.

    This implementation evaluates the elements of the iterable lazily.
    """
    if n > 0:
        saved = []
        for element in iterable:
            yield element
            saved.append(element)
        yield from chain.from_iterable(repeat(saved, n - 1))


def _lazy_product(iterables: Collection[Iterable[T]], lengths: Collection[int]) -> Iterator[Tuple[T, ...]]:
    """Return the Cartesian Product of the iterables of the given lengths.

    This implementation takes advantage of the known lengths of the iterables
    to evaluate the iterables lazily in contrast to ``itertools.product``.
    This function generates tuples in the same order as ``itertools.product``.

    Preconditions: ``len(iterables) == len(lengths)`` and ``lengths[i]`` is the
    number of calls of `next` on an iterator of `iterables[i]` before
    ``StopIteration`` is raised.
    """
    cycle_count = 1
    repeat_count = prod(lengths)
    iterators = []
    for iterable, length in zip(iterables, lengths):
        repeat_count //= length
        iterators.append(_ncycles(_repeat_each(iterable, repeat_count), cycle_count))
        cycle_count *= length
    return zip(*iterators)


class Disambiguator(Interpreter):
    """An Interpreter that iterates over the unambiguous derivations of an
    ambiguous tree (one containing '_ambig' nodes).

    By lazily constructing trees, ``Disambiguator`` is more computationally and
    memory efficient than ``lark.visitors.CollapseAmbiguities``.

    When visiting a ``CountedTree``, ``Disambiguator`` takes advantage of the
    known derivation counts to be even more lazy and is ideal for the case in
    which you only need to find one tree that meets your requirements.

    If you are always going to iterate over all possible derivations, it is
    slightly faster to visit a regular ``Tree``.
    """
    def _ambig(self, tree: Tree) -> Iterator[Tree]:
        for child in tree.children:
            yield from self.visit(child)

    def __default__(self, tree: Tree) -> Iterator[Tree]:
        if isinstance(tree, CountedTree) and tree.derivation_count == 1:
            yield tree
        else:
            yield from self._generate_subtrees(tree)

    def _generate_subtrees(self, tree: Tree) -> Iterator[Tree]:
        sub_tree_iterators = [self.visit(child) if isinstance(child, Tree) else (child,) for child in tree.children]
        if isinstance(tree, CountedTree):
            derivation_counts = [_get_derivation_count(child) for child in tree.children]
            children_lists = _lazy_product(sub_tree_iterators, derivation_counts)
        else:
            children_lists = product(*sub_tree_iterators)
        for children_list in children_lists:
            yield Tree(tree.data, list(children_list))



from __future__ import annotations

import argparse
import os
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Optional, TextIO, Union

from grascii.dictionary import build, install
from grascii.dictionary import list as list_dict
from grascii.dictionary import uninstall
from grascii.dictionary.common import (
    BUILTINS_PACKAGE,
    INSTALLATION_DIR,
    DictionaryNotFound,
    get_dictionary_installed_name,
    get_dictionary_path_name,
    is_dictionary_installed_name,
)
from grascii.dictionary.install import install_dictionary  # noqa: F401
from grascii.dictionary.list import get_built_ins, get_installed  # noqa: F401
from grascii.dictionary.uninstall import uninstall_dictionary  # noqa: F401

description = "Create and manage Grascii dictionaries"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    argparser.set_defaults(func=list_dict.cli_list)
    subparsers = argparser.add_subparsers(title="subcommands")

    build_parser = subparsers.add_parser(
        "build", description=build.description, help=build.description, aliases=["b"]
    )
    build.build_argparser(build_parser)
    build_parser.set_defaults(func=build.cli_build)

    install_parser = subparsers.add_parser(
        "install",
        description=install.description,
        help=install.description,
        aliases=["i"],
    )
    install.build_argparser(install_parser)
    install_parser.set_defaults(func=install.cli_install)

    uninstall_parser = subparsers.add_parser(
        "uninstall",
        description=uninstall.description,
        help=uninstall.description,
        aliases=["u"],
    )
    uninstall.build_argparser(uninstall_parser)
    uninstall_parser.set_defaults(func=uninstall.cli_uninstall)

    list_parser = subparsers.add_parser(
        "list",
        description=list_dict.description,
        help=list_dict.description,
        aliases=["l"],
    )
    list_dict.build_argparser(list_parser)
    list_parser.set_defaults(func=list_dict.cli_list)


class DictionaryType(Enum):
    BUILTIN = 0
    INSTALLED = 1
    LOCAL = 2


class Dictionary:
    """
    A class that represents a grascii dictionary and provides methods for reading
    entries from the dictionary.
    """

    def __init__(
        self, path: os.PathLike, dtype: DictionaryType, name: Optional[str] = None
    ) -> None:
        self.path = path
        self.name = name if name is not None else path.name
        if dtype is DictionaryType.BUILTIN or dtype is DictionaryType.INSTALLED:
            self.name = get_dictionary_installed_name(self.name)
        self.type = dtype

    def open(self, name: str) -> TextIO:
        """Open a file from the dictionary with the given name for reading.
        The caller is responsible for closing the file.

        :param name: The name of the file to open.
        :type name: str

        :returns: A text stream.
        """
        return Path(self.path, name).open()

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Dictionary) and self.path == other.path

    def __hash__(self) -> int:
        return hash(self.path)

    @classmethod
    def new(cls, name: Union[str, os.PathLike]) -> Dictionary:
        """Create a new dictionary from its installed name or a file path.

        :param name: The name of an installed dictionary (starting with ':') or \
                a path to a dictionary.
        :type name: Union[str, os.PathLike]

        :returns: A new Dictionary for the given name.
        """
        if isinstance(name, str) and is_dictionary_installed_name(name):
            if sys.version_info >= (3, 9):
                from importlib.resources import files
            else:
                from importlib_resources import files
            path_name = get_dictionary_path_name(name)
            dictionary_path = files(BUILTINS_PACKAGE).joinpath(path_name)
            if dictionary_path.is_dir():
                return Dictionary(dictionary_path, DictionaryType.BUILTIN)
            dictionary_path = INSTALLATION_DIR / path_name
            if dictionary_path.is_dir():
                return Dictionary(dictionary_path, DictionaryType.INSTALLED)
        dictionary_path = Path(name)
        if dictionary_path.is_dir():
            return Dictionary(dictionary_path, DictionaryType.LOCAL)
        raise DictionaryNotFound(name)



#!/usr/bin/env python

from __future__ import annotations

import argparse
import signal
import sys

from grascii import APP_NAME, __version__, config, dictionary, search


def main() -> None:
    try:
        from pytest_cov.embed import cleanup_on_signal
    except ImportError:
        pass
    else:
        cleanup_on_signal(signal.SIGHUP)

    argparser = argparse.ArgumentParser(prog=APP_NAME)
    subparsers = argparser.add_subparsers(title="subcommands")
    argparser.set_defaults(func=None)

    argparser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )

    search_parser = subparsers.add_parser(
        "search", description=search.description, help=search.description, aliases=["s"]
    )
    search.build_argparser(search_parser)
    search_parser.set_defaults(func=search.cli_search)

    dictionary_parser = subparsers.add_parser(
        "dictionary",
        description=dictionary.description,
        help=dictionary.description,
        aliases=["d"],
    )
    dictionary.build_argparser(dictionary_parser)

    config_parser = subparsers.add_parser(
        "config",
        description=config.description,
        help=config.description,
        aliases=["cf"],
    )
    config.build_argparser(config_parser)
    config_parser.set_defaults(func=config.cli_config)

    args = argparser.parse_args(sys.argv[1:])

    if args.func:
        args.func(args)
    else:
        argparser.print_help()


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
import sys
from pathlib import Path
from shutil import rmtree

from grascii.dictionary.common import (
    INSTALLATION_DIR,
    DictionaryNotFound,
    get_dictionary_path_name,
)

description = "Uninstall a Grascii Dictionary"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    argparser.add_argument(
        "name", action="store", help="The name of the dictionary to uninstall"
    )
    argparser.add_argument(
        "-f", "--force", action="store_true", help="Force removal of a dictionary"
    )


def uninstall_dictionary(name: str, install_dir: Path, force: bool = False) -> None:
    """Uninstall a dictionary from an installation directory.

    :param name: The name of the dictionary to uninstall.
    :param install_dir: A path to uninstall the dictionary from.
    :param force: If True, forces the uninstallation of a dictionary even in the \
            case of corruption.
    :type name: Path
    :type install_dir: Path
    :type force: bool

    """
    dictionary_path = install_dir / get_dictionary_path_name(name)
    if not dictionary_path.exists():
        raise DictionaryNotFound(get_dictionary_path_name(name))
    if force:
        rmtree(dictionary_path)
    else:
        for f in dictionary_path.iterdir():
            f.unlink()
        dictionary_path.rmdir()


def cli_uninstall(args: argparse.Namespace) -> None:
    try:
        uninstall_dictionary(args.name, INSTALLATION_DIR, force=args.force)
    except DictionaryNotFound:
        print(args.name, "does not exist.", file=sys.stderr)
    except OSError:
        print("An error occurred during uninstallation.", file=sys.stderr)
        print("The dictionary may be corrupted.", file=sys.stderr)
        print("To force removal, run with --force.", file=sys.stderr)
    else:
        print("Successfully uninstalled", args.name)


def main() -> None:
    """Run the uninstall command using arguments from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_uninstall(args)


if __name__ == "__main__":
    main()



"""Contains functions for working with a grascii configuration file.
Acts as the entry point for the grascii config command.

This can be invoked as a standalone program:
$ python -m grascii.config --help
"""

from __future__ import annotations

import argparse
import sys
from pathlib import Path, PurePath

from platformdirs import user_config_dir

from grascii import APP_NAME

CONF_DIRECTORY = user_config_dir(APP_NAME)
CONF_FILE_NAME = APP_NAME + ".conf"
DEFAULTS_CONF_NAME = "defaults.conf"

description = "Manage Grascii configuration"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    """Configure an ArgumentParser parser to parse the config command-line
    options

    :parse argparser: A fresh ArgumentParser to configure.
    """

    group = argparser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--init", action="store_true", help="Create a configuration file."
    )
    group.add_argument(
        "-D",
        "--delete",
        action="store_true",
        help="Delete an existing configuration file.",
    )
    group.add_argument(
        "-w",
        "--where",
        action="store_true",
        help="Print the path to the configuration file and exit.",
    )
    argparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Allow overwriting of an existing configuration file.",
    )


def get_default_config() -> str:
    """Get the text of the default configuration file.

    :returns: A string containing the default configuration.
    """
    if sys.version_info >= (3, 9):
        from importlib.resources import files
    else:
        from importlib_resources import files
    return files("grascii").joinpath(DEFAULTS_CONF_NAME).read_text()


def config_exists() -> bool:
    """Check whether the user configuration file exists.

    :returns: True if the configuration file exists.
    """

    return Path(get_config_file_path()).exists()


def get_config_file_path() -> PurePath:
    """Get a path object representing the location of the configuration
    file.

    :returns: A path object.
    """

    return PurePath(CONF_DIRECTORY, CONF_FILE_NAME)


def create_config() -> None:
    """Create a configuration file with the default settings."""

    Path(CONF_DIRECTORY).mkdir(parents=True, exist_ok=True)
    with Path(get_config_file_path()).open("w") as config:
        config.write(get_default_config())


def delete_config() -> None:
    """Delete a configuration file."""

    Path(get_config_file_path()).unlink()


def cli_config(args: argparse.Namespace) -> None:
    """Run a search using arguments parsed from the command line.

    :param args: A namespace of parsed arguments.
    """

    if args.where:
        if config_exists():
            print(get_config_file_path())
        else:
            print(
                "Configuration file does not exist. Use --init to create one.",
                file=sys.stderr,
            )
    elif args.init:
        if config_exists() and not args.force:
            print("Configuration file already exists.", file=sys.stderr)
            print(
                "If you would like to overwrite it with the default",
                "configuration, run again with --force.",
                file=sys.stderr,
            )
            return
        create_config()
        print("Configuration file created at", get_config_file_path())
    elif args.delete:
        if not config_exists():
            print("Configuration file does not exist.", file=sys.stderr)
            return
        if not args.force:
            print(
                "Are you sure you want to delete the configuration file?",
                "If so, run with --force.",
                file=sys.stderr,
            )
            return
        delete_config()
        print("Removed", get_config_file_path())


def main() -> None:
    """Run the config command using arguments from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_config(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

from enum import Enum
from typing import List, NamedTuple, Optional

from grascii import grammar
from grascii.parser import Interpretation


class Direction(Enum):
    NORTH = "N"
    EAST = "E"
    SOUTH = "S"
    WEST = "W"
    NORTH_WEST = "NW"
    NORTH_EAST = "NE"
    SOUTH_EAST = "SE"
    SOUTH_WEST = "SW"


class Curve(Enum):
    LINE = 1
    CLOCKWISE = 2
    COUNTER_CLOCKWISE = 3
    LOOP = 4


class StrokeType(NamedTuple):
    direction: Optional[Direction]
    curve: Optional[Curve]


class Stroke:
    """The building block of an ``Outline``"""

    def __init__(
        self,
        stroke: str,
        head_direction: Optional[Direction] = None,
        head_curve: Optional[Curve] = None,
        tail_direction: Optional[Direction] = None,
        tail_curve: Optional[Curve] = None,
    ) -> None:
        self.stroke = stroke
        self.head_type = StrokeType(head_direction, head_curve)
        td = tail_direction if tail_direction else head_direction
        tc = tail_curve if tail_curve else head_curve
        self.tail_type = StrokeType(td, tc)
        self.annotations: List[str] = []
        self.next: Optional[Stroke] = None
        self.prev: Optional[Stroke] = None
        self.next_consonant: Optional[Stroke] = None
        self.prev_consonant: Optional[Stroke] = None
        self.next_vowel: Optional[Stroke] = None
        self.prev_vowel: Optional[Stroke] = None
        self.next_char: Optional[Stroke] = None
        self.prev_char: Optional[Stroke] = None

    def has_annotation(self, annotation: str) -> bool:
        return annotation in self.annotations

    def has_direction_annotation(self) -> bool:
        return self.has_annotation(grammar.LEFT) or self.has_annotation(grammar.RIGHT)

    def add_annotation(self, annotation: str):
        self.annotations.insert(0, annotation)

    @classmethod
    def create(cls, stroke: str) -> "Stroke":
        if stroke in {"A", "E", "I", "A&'", "A&E"}:
            return cls(stroke, None, Curve.LOOP)
        if stroke in {"K", "G"}:
            return cls(stroke, Direction.EAST, Curve.CLOCKWISE)
        if stroke in {"R", "L", "LD"}:
            return cls(stroke, Direction.EAST, Curve.COUNTER_CLOCKWISE)
        if stroke in {"N", "M", "MN", "MM"}:
            return cls(stroke, Direction.EAST, Curve.LINE)
        if stroke in {"T", "D", "TD", "DT", "DD"}:
            return cls(stroke, Direction.NORTH_EAST, Curve.LINE)
        if stroke in {"P", "B"}:
            return cls(stroke, Direction.SOUTH_WEST, Curve.COUNTER_CLOCKWISE)
        if stroke in {"F", "V"}:
            return cls(stroke, Direction.SOUTH_WEST, Curve.CLOCKWISE)
        if stroke in {"SH", "CH", "J"}:
            return cls(stroke, Direction.SOUTH_WEST, Curve.LINE)
        if stroke in {"NT", "ND", "MT", "MD"}:
            return cls(stroke, Direction.NORTH_EAST, Curve.COUNTER_CLOCKWISE)
        if stroke in {"TN", "DN", "TM", "DM"}:
            return cls(stroke, Direction.NORTH_EAST, Curve.CLOCKWISE)
        if stroke in {"NG", "NK"}:
            return cls(stroke, Direction.SOUTH_EAST, Curve.LINE)
        if stroke in {"DF", "DV", "TV", "U"}:
            return cls(
                stroke,
                Direction.NORTH_EAST,
                Curve.CLOCKWISE,
                Direction.SOUTH_WEST,
                Curve.CLOCKWISE,
            )
        if stroke in {"PNT", "PND", "JNT", "JND", "O"}:
            return cls(
                stroke,
                Direction.SOUTH_EAST,
                Curve.COUNTER_CLOCKWISE,
                Direction.NORTH_EAST,
                Curve.COUNTER_CLOCKWISE,
            )
        return cls(stroke)


class Outline:

    """
    An alternative to ``Interpretation`` as a representation of a
    Grascii string. It is structured as a linked list and is better for
    contextual processing of strokes. Outlines infer the directions of
    directional characters and explicitly add direction annotations.
    """

    def __init__(self, interpretation: Interpretation) -> None:
        self.first = None
        self.last = None
        self._build(interpretation)
        self._infer_directions()

    def _build(self, interpretation: Interpretation) -> None:
        needs_next_consonant = []
        needs_next_vowel = []
        needs_next_char = []
        prev_stroke = None
        prev_vowel = None
        prev_consonant = None
        prev_char = None
        for item in interpretation:
            if not isinstance(item, list):
                stroke = Stroke.create(item)
                if prev_stroke:
                    prev_stroke.next = stroke
                else:
                    self.first = stroke
                stroke.prev = prev_stroke
                stroke.prev_consonant = prev_consonant
                stroke.prev_vowel = prev_vowel
                stroke.prev_char = prev_char
                if item in grammar.CONSONANTS:
                    while needs_next_consonant:
                        needs_next_consonant.pop().next_consonant = stroke
                    prev_consonant = stroke
                    while needs_next_char:
                        needs_next_char.pop().next_char = stroke
                    prev_char = stroke
                elif item in grammar.VOWELS:
                    while needs_next_vowel:
                        needs_next_vowel.pop().next_vowel = stroke
                    prev_vowel = stroke
                    while needs_next_char:
                        needs_next_char.pop().next_char = stroke
                    prev_char = stroke
                elif item == grammar.BOUNDARY or item == grammar.DISJOINER:
                    needs_next_consonant.clear()
                    needs_next_vowel.clear()
                    needs_next_char.clear()
                    prev_consonant = None
                    prev_vowel = None
                    prev_char = None
                needs_next_consonant.append(stroke)
                needs_next_vowel.append(stroke)
                needs_next_char.append(stroke)
                self.last = stroke
                prev_stroke = stroke
            else:
                prev_stroke.annotations = item.copy()

    def _infer_directions(self) -> None:
        """
        Add direction annotations to strokes without explicit directions. This
        method sets the directions of O, U, S/Z, and TH according to the
        following rules from the Preanniversary edition of Gregg Shorthand:

        20. The O-hook is placed on its side before N, M, R, L except when
        preceded by a downward character.

        22. The OO-hook is always placed on its side after N or M; it is also
        placed on its side after K or G when followed by R or L.

        30. When S is joined to a curve, S is written in the same direction as
        the curve to which it is joined. A circle vowel occurring at the joining
        does not affect the application of this rule.

        31. When S is joined to T, D, N, M, the S is used which forms a sharp
        angle. A circle vowel occurring at the joining does not affect the
        application of this rule.

        32. When S is joined to SH, CH, J the S is used which is written with
        the clockwise movement.

        33. In words consisting of S or TH, or both, and a circle vowel, S or TH
        should be written with the clockwise movement.

        34. The clockwise TH is given the preference, but when joined to O, R,
        L the other form is used.

        35. In words beginning with so, the comma S is used.

        Addendums
        =========

        The following addendums are used to address cases not covered by the
        above rules. These rules do not appear explicitly in the Gregg
        manual, but may be inferred, or simply chosen to ease the
        implementation.

        1. When S is between two consonants, its direction is based on the
        preceding character unless there is a syllable break forcing its
        direction to be based on the next character.

        2. When TH is joined to R or L, the under TH is used. A circle vowel
        occurring at the joining does not affect the application of this rule.

        3. In words beginning with IS, the left S is used.

        """

        def set_S_stroke_type(stroke: Stroke) -> None:
            assert stroke.has_direction_annotation()
            if stroke.has_annotation(grammar.RIGHT):
                stroke.head_type = StrokeType(Direction.SOUTH_WEST, Curve.CLOCKWISE)
                stroke.tail_type = StrokeType(Direction.SOUTH_WEST, Curve.CLOCKWISE)
            elif stroke.has_annotation(grammar.LEFT):
                stroke.head_type = StrokeType(
                    Direction.SOUTH_WEST, Curve.COUNTER_CLOCKWISE
                )
                stroke.tail_type = StrokeType(
                    Direction.SOUTH_WEST, Curve.COUNTER_CLOCKWISE
                )

        def set_TH_stroke_type(stroke: Stroke) -> None:
            assert stroke.has_direction_annotation()
            if stroke.has_annotation(grammar.RIGHT):
                stroke.head_type = StrokeType(
                    Direction.NORTH_EAST, Curve.COUNTER_CLOCKWISE
                )
                stroke.tail_type = StrokeType(
                    Direction.NORTH_EAST, Curve.COUNTER_CLOCKWISE
                )
            elif stroke.has_annotation(grammar.LEFT):
                stroke.head_type = StrokeType(Direction.NORTH_EAST, Curve.CLOCKWISE)
                stroke.tail_type = StrokeType(Direction.NORTH_EAST, Curve.CLOCKWISE)

        def set_S_direction(stroke: Stroke) -> None:
            assert stroke.stroke == "S" or stroke.stroke == "Z"
            if (
                not stroke.prev_char
                and stroke.next_char
                and stroke.next_char.stroke == "O"
            ):
                # Rule 35
                stroke.add_annotation(grammar.RIGHT)
                return
            # Addendum 1
            if stroke.prev_char:
                if stroke.prev_char.tail_type.curve == Curve.LOOP:
                    if stroke.prev_consonant:
                        # Rule 30 + 31
                        if set_S_direction_based_on_curves(
                            stroke, stroke.prev_consonant.tail_type, is_before=False
                        ):
                            return
                    elif (
                        stroke.prev_char.stroke == "I"
                        and not stroke.prev_char.prev_char
                    ):
                        # Addendum 3
                        stroke.add_annotation(grammar.LEFT)
                        return
                # Rule 31
                if set_S_direction_based_on_curves(
                    stroke, stroke.prev_char.tail_type, is_before=False
                ):
                    return
            if stroke.next_char:
                if stroke.next_char.head_type.curve == Curve.LOOP:
                    if stroke.next_consonant:
                        # Rule 30 + 31
                        if set_S_direction_based_on_curves(
                            stroke, stroke.next_consonant.head_type, is_before=True
                        ):
                            return
                # Rule 31
                if set_S_direction_based_on_curves(
                    stroke, stroke.next_char.head_type, is_before=True
                ):
                    return

            # Rule 33
            stroke.add_annotation(grammar.RIGHT)

        def set_S_direction_based_on_curves(
            stroke: Stroke, stroke_type: StrokeType, is_before: bool
        ) -> bool:
            if stroke_type.curve == Curve.CLOCKWISE:
                # Rule 30
                stroke.add_annotation(grammar.RIGHT)
                return True
            elif stroke_type.curve == Curve.COUNTER_CLOCKWISE:
                # Rule 30
                stroke.add_annotation(grammar.LEFT)
                return True
            elif stroke_type.curve == Curve.LINE:
                if stroke_type.direction == Direction.SOUTH_WEST:
                    # Rule 32
                    stroke.add_annotation(grammar.RIGHT)
                    return True
                else:
                    # Rule 31
                    if is_before:
                        stroke.add_annotation(grammar.RIGHT)
                    else:
                        stroke.add_annotation(grammar.LEFT)
                    return True
            return False

        def set_TH_direction(stroke: Stroke) -> None:
            assert stroke.stroke == "TH"
            if (stroke.next_char and stroke.next_char.stroke in {"O", "R", "L"}) or (
                stroke.prev_char and stroke.prev_char.stroke in {"O", "R", "L"}
            ):
                # Rule 34
                stroke.add_annotation(grammar.RIGHT)
            elif (
                stroke.next_char
                and stroke.next_char.stroke in {"A", "E", "I", "A&'", "A&E"}
                and stroke.next_consonant
                and stroke.next_consonant.stroke in {"R", "L"}
            ) or (
                stroke.prev_char
                and stroke.prev_char.stroke in {"A", "E", "I", "A&'", "A&E"}
                and stroke.prev_consonant
                and stroke.prev_consonant.stroke in {"R", "L"}
            ):
                # Addendum 2
                stroke.add_annotation(grammar.RIGHT)
            else:
                # Rule 33
                stroke.add_annotation(grammar.LEFT)

        def set_O_direction(stroke: Stroke) -> None:
            assert stroke.stroke == "O"
            if stroke.next_char and stroke.next_char.stroke in {
                "N",
                "M",
                "MN",
                "MM",
                "R",
                "L",
            }:
                if (not stroke.prev) or (
                    not stroke.prev_char
                    or stroke.prev_char.tail_type.direction != Direction.SOUTH_WEST
                ):
                    # Rule 20
                    stroke.add_annotation(grammar.LEFT)

        def set_U_direction(stroke: Stroke) -> None:
            assert stroke.stroke == "U"
            if stroke.prev_char:
                if stroke.prev_char.stroke in {"N", "M", "MN", "MM"}:
                    # Rule 22
                    stroke.add_annotation(grammar.RIGHT)
                elif stroke.prev_char.stroke in {"K", "G"}:
                    if stroke.next_char and stroke.next_char.stroke in {"R", "L"}:
                        # Rule 22
                        stroke.add_annotation(grammar.RIGHT)

        current_stroke = self.first
        while current_stroke:
            if current_stroke.stroke in {"S", "Z"}:
                if not current_stroke.has_direction_annotation():
                    set_S_direction(current_stroke)
                set_S_stroke_type(current_stroke)
            elif current_stroke.stroke == "TH":
                if not current_stroke.has_direction_annotation():
                    set_TH_direction(current_stroke)
                set_TH_stroke_type(current_stroke)
            elif current_stroke.stroke == "O":
                if not current_stroke.has_direction_annotation():
                    set_O_direction(current_stroke)
            elif current_stroke.stroke == "U":
                if not current_stroke.has_direction_annotation():
                    set_U_direction(current_stroke)
            current_stroke = current_stroke.next

    def __str__(self) -> str:
        builder = []
        current_stroke = self.first
        while current_stroke:
            builder.append(current_stroke.stroke)
            builder += current_stroke.annotations
            current_stroke = current_stroke.next
        return "".join(builder)

    def to_interpretation(self) -> Interpretation:
        """Convert this ``Outline`` to an ``Interpretation``"""
        interpretation: Interpretation = []
        current_stroke = self.first
        while current_stroke:
            interpretation.append(current_stroke.stroke)
            if current_stroke.annotations:
                interpretation.append(current_stroke.annotations)
            current_stroke = current_stroke.next
        return interpretation
